#include <stdio.h>
#include <conio.h>
#include <windows.h>

DWORD find_torrent_data(HANDLE hFile, int *len) 
{
	DWORD bytesread;
	IMAGE_DOS_HEADER dos_header;
	IMAGE_NT_HEADERS nt_header;
	IMAGE_SECTION_HEADER objtbl_entry;
	int i;
	DWORD retVal;

	int sig, torrent_size;

	*len = 0;

	// First try and look for the offset written at the end of the file
	// The last 8 bytes of the file should be that 4 byte signature "PSER"
	// followed by a 4byte signed int offset to the start of the torrent
	// from the start of the sig
	SetFilePointer(hFile, -8, 0, SEEK_END);
	ReadFile(hFile, &sig, 4, &bytesread, NULL);
	if (bytesread == 4 && sig == 0x52455350)
	{
		ReadFile(hFile, &torrent_size, 4, &bytesread, NULL);
		if (bytesread == 4)
		{
			*len = -torrent_size;
			return SetFilePointer(hFile, 0, 0, SEEK_END) + torrent_size - 8;
		}
	}

	// If not found that way, we fall back to trying to find any data that's described in the
	// PE section headers
	SetFilePointer(hFile, 0, 0, SEEK_SET);

	ReadFile(hFile, &dos_header, sizeof(dos_header), &bytesread, NULL);
	if (bytesread != sizeof(dos_header) || dos_header.e_magic != IMAGE_DOS_SIGNATURE)
		return 0;

	SetFilePointer(hFile, dos_header.e_lfanew, NULL, FILE_BEGIN);
	ReadFile(hFile, &nt_header, sizeof(nt_header), &bytesread, NULL);
	if (bytesread != sizeof(nt_header) || nt_header.Signature != IMAGE_NT_SIGNATURE ||
		nt_header.FileHeader.SizeOfOptionalHeader == 0) 
		return 0;

	retVal = 0;
	for (i=0; i<nt_header.FileHeader.NumberOfSections; i++) {
		ReadFile(hFile, &objtbl_entry, sizeof(IMAGE_SECTION_HEADER), &bytesread, NULL);
		if (objtbl_entry.PointerToRawData >= retVal)
			retVal = objtbl_entry.PointerToRawData + objtbl_entry.SizeOfRawData;
	}  /* for i -> numsections */
	
	return retVal;
}

BOOL copy_file_part(HANDLE hFile, const char *fnameout, const DWORD start, DWORD len)
{
	char buff[4096];
	DWORD dwRead;
	HANDLE hFileOut;

	if ((hFileOut = CreateFile(fnameout, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, 
		NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, 0)) == INVALID_HANDLE_VALUE)
		return FALSE;

	SetFilePointer(hFile, start, 0, FILE_BEGIN);

	while (len)
	{
		dwRead = 0;
		if (len > sizeof(buff))
			ReadFile(hFile, buff, sizeof(buff), &dwRead, NULL);
		else
			ReadFile(hFile, buff, len, &dwRead, NULL);
		if (dwRead)
			WriteFile(hFileOut, buff, dwRead, &dwRead, NULL);
		else
			break;
		len -= dwRead;
	}

	CloseHandle(hFileOut);
	return TRUE;
}

static const char *DEFAULT_SEARCH_PATH = "WoW-*-downloader.exe";

int main(int argc, char *argv[])
{
	WIN32_FIND_DATA wfd;
	HANDLE hFind;
	HANDLE hFile;
	DWORD start, len;
	FILETIME highestft;
	char fname[MAX_PATH], fnameout[MAX_PATH];
	char *fnamedot, *searchpathslash;
	const char *searchpath;

	if (argc > 1)
		searchpath = argv[1];
	else
		searchpath = DEFAULT_SEARCH_PATH;

	if ((hFind = FindFirstFile(searchpath, &wfd)) == INVALID_HANDLE_VALUE)
	{
		printf("Could not find any %s\n", searchpath);
		printf("Press any key to exit..\n");
		getch();
		return 1;
	}

	ZeroMemory(&highestft, sizeof(highestft));
	do {
		if (wfd.ftLastWriteTime.dwHighDateTime > highestft.dwHighDateTime || 
			(wfd.ftLastWriteTime.dwHighDateTime = highestft.dwHighDateTime && wfd.ftLastWriteTime.dwLowDateTime > highestft.dwLowDateTime)
			)
		{
			highestft = wfd.ftLastWriteTime;
			strcpy(fname, wfd.cFileName);
		}

		if (FindNextFile(hFind, &wfd) == 0) 
			break;
	} while (NULL != hFind && INVALID_HANDLE_VALUE != hFind);

	FindClose(hFind);

	// If the search path contains a path delimiter, it probably has a path 
	// we need to prepend to the filename.  Used in the case someone passes
	// a path like C:\Games\WoW-*.exe or someone dragged a file onto us
	searchpathslash = strrchr(searchpath, '\\');
	if (searchpathslash)
	{
		// use the out buffer temporarily to hold the found file name
		strcpy(fnameout, fname);
		strncpy(fname, searchpath, searchpathslash - searchpath + 1);
		fname[searchpathslash - searchpath + 1] = 0;
		strcat(fname, fnameout);
	}

	printf("Extracting from %s...\n", fname);

	if ((hFile = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, 
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_ARCHIVE, 0)) == INVALID_HANDLE_VALUE)  
	{
		char dirbuff[MAX_PATH];
		printf("Could not open input file!  Error %u\n", GetLastError());
		if (GetCurrentDirectory(sizeof(dirbuff), dirbuff))
			printf("  current directory %s\n", dirbuff);
		printf("Press any key to exit..\n");
		getch();
		return 1;
	}

	Sleep(1000);
	if (start = find_torrent_data(hFile, &len))
	{
		if (len == 0)
			len = SetFilePointer(hFile, 0, 0, FILE_END) - start;
		printf("Torrent should be at 0x%x len 0x%x\n", start, len);

		fnamedot = strrchr(fname, '.');
		if (fnamedot)
		{
			*fnamedot = 0;
			sprintf(fnameout, "%s.torrent", fname);
			printf("Saving torrent to %s...\n", fnameout);
			if (copy_file_part(hFile, fnameout, start, len))
				ShellExecute(0, NULL, fnameout, NULL, NULL, SW_NORMAL);
		}
	}  /* if found start */
	
	CloseHandle(hFile);

	return 0;
}